/***************************************************************
 **   This is the module that obtains a NxN upper triangular  **
 ** matrix, from a non-singular NxN matrix. Also solveU,      **
 ** solves the linear equation defined by this upper          **
 ** triangular matrix.                                        **
 **                                                           **
 **                   Written by Emre Tuncer                  **
 **                    6.1990, METU, Ankara                   **
 **              Last Modified: 3.1996, Sangwoo Kim           **
 ** Modification includes new memory allocation scheme and    **
 ** LU decomposition                                          ** 
 ***************************************************************/
#include <stream.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <stddef.h>
#include "../src/complex.h"


double** allocator(int N)
{
  int i;
  double** A=new double*[N];

  for (i=0;i<N;i++) A[i]=new double[N];
  return A;
}


void delocator(int N,double** A)
{
  int i;
  for (i=0;i<N;i++) delete A[i];
}

complex **matrix(int n,int m)
{
  complex **p;
  int i;

  p=new complex*[n];
  if (p==(complex**) NULL) {
    perror("matrix: Out of memory");
    exit(1);
  }
  for (i=0;i<n;i++) {
    p[i]=new complex[m];
    if (p[i]==(complex*) NULL) {
      perror("matrix: Out of memory");
      exit(1);
    } 
  }
  return p;
}

void mfree(complex** p,int n)
{
  int i;
  
  for (i=0;i<n;i++) delete p[i];
  delete p;
  return;
}

double **dmatrix(int n,int m)
{
  double **p;
  int i;

  p=new double*[n];
  if (p==(double**) NULL) {
    perror("dmatrix: Out of memory");
    exit(1);
  }
  for (i=0;i<n;i++) {
    p[i]=new double[m];
    if (p[i]==(double*) NULL) {
      perror("dmatrix: Out of memory");
      exit(1);
    }
  }
  return p;
}

void dmfree(double** p,int n)
{
  int i;

  for (i=0;i<n;i++) delete p[i];
  delete p;
  return;
}



void toU(complex** B,int N,complex** D,int M)
{
  int i,j,l;
  double maxr;
  complex temp,coef,sum;


  for (i=0;i<N;i++) {
    maxr=0.0;
    for (j=i;j<N;j++)
      if (cabs_(B[j][i])>maxr) {
        maxr=cabs_(B[j][i]);
        l=j;
      }
    if (maxr==0.0) {
      perror("\nSingular Matrix Error **\n");
      return;
    }
    for (j=i;j<N;j++) {
      temp=B[i][j];
      B[i][j]=B[l][j];
      B[l][j]=temp;
    }
    for (j=0;j<M;j++) {
      temp=D[i][j];
      D[i][j]=D[l][j];
      D[l][j]=temp;
    }
    for (j=i+1;j<N;j++) {
      coef=B[j][i]/B[i][i];
      for (l=i;l<N;l++) B[j][l]=B[j][l]-coef*B[i][l];
      for (l=0;l<M;l++) D[j][l]=D[j][l]-coef*D[i][l];
    }
  }
}


complex** solveU(complex** B,int N,complex** D,int M)
{
  int i,j,l;
  complex sum;
  complex** X;

  X=matrix(N,M);

  for (l=0;l<M;l++) {
    X[N-1][l]=D[N-1][l]/B[N-1][N-1];
    for (i=N-2;i>=0;i--)  {
      sum=r2c(0.,0.);
      for (j=i+1;j<N;j++) sum=sum+X[j][l]*B[i][j];
      X[i][l]=(D[i][l]-sum)/B[i][i];
    }
  }
  mfree(D,N);
  return X;
}


  

complex det(complex** A,int N)
{
  complex** D;
  complex dd=r2c(1.0,0.);
  int i;  
  
  D=matrix(N,1);
  for (i=0;i<N;i++) D[i][0]=r2c(0.,0.);
  toU(A,N,D,1);
  for (i=0;i<N;i++) dd=dd*A[i][i];
  return dd;
  mfree(D,N);
}



complex** inverse(complex** A,int N)
{
  complex** D;
  int i,j; 
  D=matrix(N,N);
  for (i=0;i<N;i++) {
    for ( j=0;j<N;j++) {
      D[i][j]=r2c(0.,0.);
    }
  }
  for (i=0;i<N;i++) D[i][i]=r2c(1.0,0.);
  toU(A,N,D,N);
  D=solveU(A,N,D,N);
  mfree(A,N);
  return D;
}

